/* obstacle.h
   Stephen Sorkin
   This file defines the Obstacle class.
   An instance of the Obstacle class is a convenient way to represent a 2D
   obstacle on a planar workspace.  The Obstacle is stored as a series of
   vertices.  The vertices are stored by reference number.  The Obstacle's
   workspace must be queried whenever the exact Point2 must be known.  The
   Obstacle must have its vertices specified in counterclockwise order.
   */

#ifndef _obstacle_h
#define _obstacle_h

#include <iostream.h>
#include "distance/dist.h"

class Workspace;

class Obstacle {
public:
  Obstacle():work(NULL), nvert(0), vertices(NULL) {}
  Obstacle(istream& in, Workspace *myWork);

  Sobj2 *getQuinRep() const {return quinRep;}
  const bool pointInside(const Point2& point) const;
  //double getPointDist(Frame2 *pointFrame) const;
  double getMaxPen() const {return maxpen;}

private:
  Workspace *work;

  int nvert;
  int *vertices;

  double maxpen;

  mutable Sobj2 *quinRep;

  void readObstData(istream& obstData);
  void buildQuinRep();
};

#endif
