/* pathnode.h
   Stephen Sorkin
   This file defines the PathNode class.
   */

#ifndef _pathnode_h
#define _pathnode_h

#include "darray.h"
#include "workspace.h"
#include "obstacle.h"

class PathEdge;
class PathPlanner;

class PathNode {
 public:
  static PathNode *getNodePtr(const Point2& point, double penFrac, const Workspace *workspace);
  ~PathNode();

  bool moveToFree(double newPenFrac);

  double getDist() const {return nodeState.dist;}
  double getPenFrac() const {return (nodeState.dist/nodeState.obst->getMaxPen());}
  double getMaxPen() const {return nodeState.obst->getMaxPen();}
  Point2 getPos() const {return nodeFrame.p;}

  double distToNode(const PathNode *next) const
    {return pabs2(psub2(nodeFrame.p, next->nodeFrame.p));}

  int numConnections() const {return connections.length();}

  void addNeighbor(PathNode *node, bool primary)
    {primary ? pNeighbors.append(node) : sNeighbors.append(node);}
  void removeNeighbor(PathNode *node, bool primary)
    {primary ? pNeighbors.remove(node) : sNeighbors.remove(node);}

  void findEdges(double maxPenFrac, PathPlanner *pathPlanner);
  void addEdge(PathEdge *edge, PathNode *other);
  void removeEdge(PathEdge *edge, PathNode *other);

  DArray<PathEdge *>& getEdges() {return connections;}

  static void saveNodesVect(DArray<PathNode *>& nodes, const char *outFileName)
;
  static int getNumCalls() {return numCalls;}

 private:
  Frame2 nodeFrame;
  RobotState nodeState;
  const Workspace *workspace;

  DArray<PathNode *> pNeighbors;
  DArray<PathNode *> sNeighbors;
  DArray<PathEdge *> connections;
  DArray<PathNode *> connNodes;

  PathNode(const Frame2& myNodeFrame, const RobotState& myNodeState, 
	   const Workspace *myWorkspace):
    nodeFrame(myNodeFrame), nodeState(myNodeState), workspace(myWorkspace) {}

  static int numCalls;
};

#endif
