/* pathplanner.h
   Stephen Sorkin
   This file defines the Pathplanner class.
   */

#ifndef _pathplanner_h
#define _pathplanner_h

#include <iostream.h>
#include "darray.h"
#include "workspace.h"

class PathEdge;
class PathNode;

class PathPlanner {
 public:
  PathPlanner(istream& workspaceIn, double myInitPenetration);
  ~PathPlanner();

  void makePenNodes(int initNumNodes);
  void makePenPaths(double proximity);

  void findNeighbors(double proximity);

  void regrowObstacles(double penDecr);

  void saveNodesVect(const char *outFileName);
  void savePathsVect(const char *outFileName);

  void addEdge(PathEdge *edge) {assert(edge); pathEdges.append(edge);}
  void removeEdge(PathEdge *edge) {pathEdges.remove(edge);}

  void addNode(PathNode *node) {assert(node); pathNodes.append(node);}
  void removeEdge(PathNode *node) {pathNodes.remove(node);}

  bool connected();

 private:
  Workspace workspace;

  double initPen, curPen;
  
  DArray<PathNode *> pathNodes;
  DArray<PathEdge *> pathEdges;

  PathNode *tStart, *tEnd;
};

#endif

